package com.dzeio.openhealth.connectors.samsunghealth

import android.app.Activity
import android.app.AlertDialog
import android.util.Log
import com.samsung.android.sdk.healthdata.*

class SamsungHealth(
    private val activity: Activity
) {

    private var currentStartTime: Long = 0
    private var isFinishing = false

    init {
        currentStartTime = StepCountReader.TODAY_START_UTC_TIME
    }

    companion object {
        const val TAG = "SamsungHealth"
    }

//    private val binningListAdapter: StepBinningData by lazy { StepBinningData() }

    private val healthDataStore: HealthDataStore by lazy { HealthDataStore(activity, connectionListener) }
    private val stepCountReader: StepCountReader by lazy { StepCountReader(healthDataStore, stepCountObserver) }

    private val stepCountObserver: StepCountObserver = object : StepCountObserver {
        override fun onChanged(count: Int) {
            Log.d(TAG, "$count")
        }

        override fun onBinningDataChanged(binningCountList: List<StepBinningData>) {
            Log.d(TAG, "${binningCountList.size}")
        }
    }

    private val connectionListener: HealthDataStore.ConnectionListener = object : HealthDataStore.ConnectionListener {
        override fun onConnected() {
            Log.d(TAG, "onConnected")
            if (checkPermissionsAcquired()) {
                stepCountReader.requestDailyStepCount(currentStartTime)
            } else {
                requestPermission()
            }
        }

        override fun onConnectionFailed(error: HealthConnectionErrorResult) {
            Log.d(TAG, "onConnectionFailed")
            showConnectionFailureDialog(error)
        }

        override fun onDisconnected() {
            Log.d(TAG, "onDisconnected")
            if (!isFinishing) {
                healthDataStore.connectService()
            }
        }
    }

    private fun showConnectionFailureDialog(error: HealthConnectionErrorResult) {
        if (isFinishing) {
            return
        }
        val alert = AlertDialog.Builder(activity)
        if (error.hasResolution()) {
            when (error.errorCode) {
                HealthConnectionErrorResult.PLATFORM_NOT_INSTALLED -> alert.setMessage("R.string.msg_req_install")
                HealthConnectionErrorResult.OLD_VERSION_PLATFORM -> alert.setMessage("R.string.msg_req_upgrade")
                HealthConnectionErrorResult.PLATFORM_DISABLED -> alert.setMessage("R.string.msg_req_enable")
                HealthConnectionErrorResult.USER_AGREEMENT_NEEDED -> alert.setMessage("R.string.msg_req_agree")
                else -> alert.setMessage("R.string.msg_req_available")
            }
        } else {
            alert.setMessage("R.string.msg_conn_not_available")
        }
        alert.setPositiveButton("R.string.ok") { _, _ ->
            if (error.hasResolution()) {
                error.resolve(activity)
            }
        }
        if (error.hasResolution()) {
            alert.setNegativeButton("R.string.cancel", null)
        }
        alert.show()
    }

    // Check whether the permissions that this application needs are acquired
    private fun checkPermissionsAcquired(): Boolean {
        val pmsManager = HealthPermissionManager(healthDataStore)

        // Check whether the permissions that this application needs are acquired
        return runCatching { pmsManager.isPermissionAcquired(permissionKeySet) }
            .onFailure { Log.e(TAG, "Permission request fails.", it) }
            .map { it.values.all { it } }
            .getOrDefault(false)
    }

    private fun requestPermission() {
        val pmsManager = HealthPermissionManager(healthDataStore)

        // Show user permission UI for allowing user to change options
        runCatching { pmsManager.requestPermissions(permissionKeySet, activity) }
            .onFailure { Log.e(TAG, "Permission setting fails.", it) }
            .getOrNull()
            ?.setResultListener(mPermissionListener)
    }

    private val permissionKeySet: Set<HealthPermissionManager.PermissionKey> =
        setOf(
            HealthPermissionManager.PermissionKey(HealthConstants.StepCount.HEALTH_DATA_TYPE, HealthPermissionManager.PermissionType.READ),
            HealthPermissionManager.PermissionKey(
                HealthConstants.StepDailyTrend.HEALTH_DATA_TYPE,
                HealthPermissionManager.PermissionType.READ
            )
        )

    private val mPermissionListener = HealthResultHolder.ResultListener<HealthPermissionManager.PermissionResult> { result ->
        // Show a permission alarm and clear step count if permissions are not acquired
        if (result.resultMap.values.any { !it }) {
            // List is now empty
            showPermissionAlarmDialog()
        } else {
            // Get the daily step count of a particular day and display it
            stepCountReader.requestDailyStepCount(currentStartTime)
        }
    }

    private fun showPermissionAlarmDialog() {
        if (isFinishing) {
            return
        }
        AlertDialog.Builder(activity)
            .setTitle("R.string.notice")
            .setMessage("R.string.msg_perm_acquired")
            .setPositiveButton("R.string.ok", null)
            .show()
    }

    fun importStepCount() {
        healthDataStore.connectService()
    }
}